import numpy as np

from .error import RunError


class Tile:
    def __init__(self, opset_version, **kwargs):
        self.version = opset_version

    def run(self, x, repeats, axis=None):
        if self.version >= 6:
            if np.ndim(x) == repeats.size:
                return [np.tile(x, repeats)]
            else:
                raise RunError("Tile", self.version)
        else:
            raise RunError("Tile", self.version)



import numpy as np


class Gather:
    def __init__(self, opset_version, **kwargs):
        self.version = opset_version
        self.axis = kwargs.get("axis", 0)

    def run(self, data, indices):
        return [np.take(data, indices, axis=self.axis)]



import numpy as np


class ScatterND:
    def __init__(self, opset_version, **kwargs):
        self.version = opset_version
        self.reduction = kwargs.get("reduction", "none")

    def run(self, data, indices, updates):
        return [scatter_nd_impl(data, indices, updates, self.reduction)]


# The following code has been copied from
# https://github.com/onnx/onnx/blob/07c494bf077e9e4a7898119f28a50585469ad4cd/onnx/backend/test/case/node/scatternd.py#L15-L31
# Copyrights (c) ONNX Project Contributers
# License: Apache-2.0
# NOTE: Modify i -> tuple(i), indices[i] -> tuple(indices[i])
def scatter_nd_impl(data, indices, updates, reduction="none"):  # type: ignore

    # Check tensor shapes
    assert indices.shape[-1] <= len(data.shape)
    assert updates.shape == indices.shape[:-1] + data.shape[indices.shape[-1] :]

    # Compute output
    output = np.copy(data)
    for i in np.ndindex(indices.shape[:-1]):
        # NOTE: The order of iteration in this loop is not specified.
        if reduction == "add":
            output[tuple(indices[i])] += updates[tuple(i)]
        elif reduction == "mul":
            output[tuple(indices[i])] *= updates[tuple(i)]
        else:
            output[tuple(indices[i])] = updates[tuple(i)]
    return output



import numpy as np


class DequantizeLinear:
    def __init__(self, opset_version, **kwargs):
        self.version = opset_version
        self.axis = kwargs.get("axis", 1)

    def run(self, x, x_scale, x_zero_point=None):
        if x_zero_point is None:
            z = np.zeros(x_scale.size).reshape(x_scale.shape).astype(np.uint8)
        else:
            z = x_zero_point

        if x_scale.size == 1:
            y = (x.astype(np.float32) - z.astype(np.float32)) * x_scale
            return [y]

        else:
            sh = np.ones(len(x.shape)).astype(int)
            sh[self.axis] = x_scale.size
            y = (x.astype(np.float32) - z.reshape(sh).astype(np.float32)) * x_scale.reshape(sh)
            return [y]



import numpy as np

from .error import RunError


class TopK:
    def __init__(self, opset_version, **kwargs):
        self.version = opset_version
        self.k = kwargs.get("k")
        self.axis = kwargs.get("axis", -1)
        self.sorted = kwargs.get("sorted", 1)
        self.largest = kwargs.get("largest", 1)

    def run(self, x, k=None):
        if self.version >= 10:
            self.k = k

        if self.k is None:
            raise RunError("TopK", self.version)

        values, indices = topk_sorted_implementation(x, self.k, self.axis, self.largest)
        return [values, indices]


# The following code has been copied from
# https://github.com/onnx/onnx/blob/2875f51853133d9cc028bbdccb62f92745cb94c2/onnx/backend/test/case/node/topk.py#L15-L23
# Copyrights (c) ONNX Project Contributers
# License: Apache-2.0
def topk_sorted_implementation(X, k, axis, largest):  # type: ignore
    sorted_indices = np.argsort(X, axis=axis)
    sorted_values = np.sort(X, axis=axis)
    if largest:
        sorted_indices = np.flip(sorted_indices, axis=axis)
        sorted_values = np.flip(sorted_values, axis=axis)
    topk_sorted_indices = np.take(sorted_indices, np.arange(k), axis=axis)
    topk_sorted_values = np.take(sorted_values, np.arange(k), axis=axis)
    return topk_sorted_values, np.array(topk_sorted_indices, dtype=np.int64)



import numpy as np


class Transpose:
    def __init__(self, opset_version, **kwargs):
        self.version = opset_version
        self.perm = kwargs.get("perm")

    def run(self, x):
        y = x.copy()
        return [np.transpose(y, axes=self.perm)]



import numpy as np


class Flatten:
    def __init__(self, opset_version, **kwargs):
        self.version = opset_version
        self.axis = kwargs.get("axis", 1)

    def run(self, x):
        shape = x.shape
        assert len(shape) >= self.axis
        new_shape = (np.prod(shape[0 : self.axis]).astype(int), -1)
        return [np.reshape(x, new_shape)]



import numpy as np


class ReduceMin:
    def __init__(self, opset_version, **kwargs):
        self.version = opset_version
        self.axes = kwargs.get("axes")
        self.keepdims = kwargs.get("keepdims", 1)

    def run(self, data):
        if self.axes is None:
            return [np.minimum.reduce(data, axis=self.axes, keepdims=self.keepdims == 1)]
        else:
            return [np.minimum.reduce(data, axis=tuple(self.axes), keepdims=self.keepdims == 1)]



import numpy as np


class Dropout:
    def __init__(self, opset_version, **kwargs):
        self.version = opset_version
        self.seed = kwargs.get("seed")
        self.consumed_inputs = kwargs.get("consumed_inputs")
        self.is_test = kwargs.get("is_test", 0)
        self.ratio = kwargs.get("ratio", 0.5)

    def run(self, data, ratio=None, training_mode=None):
        if self.version >= 12:
            if ratio is not None:
                self.ratio = ratio
            if training_mode is not None:
                return dropout(data, self.ratio, self.seed, training_mode.item(), return_mask=True)
            else:
                return dropout(data, self.ratio, self.seed, return_mask=True)
        else:
            return [data, np.zeros(data.shape, dtype=bool)]


# The following code has been copied from
# https://github.com/onnx/onnx/blob/91bf3f52c96919af8de4d90c4e70316a47112178/onnx/backend/test/case/node/dropout.py#L17-L30
# Copyrights (c) ONNX Project Contributers
# License: Apache-2.0
def dropout(X, drop_probability=0.5, seed=0, training_mode=False, return_mask=False):  # type: ignore
    if drop_probability == 0 or training_mode is False:
        if return_mask is True:
            return X, np.ones(X.shape, dtype=bool)
        else:
            return X

    np.random.seed(seed)
    mask = np.random.uniform(0, 1.0, X.shape) >= drop_probability
    scale = 1 / (1 - drop_probability)
    if return_mask is True:
        return mask * X * scale, mask.astype(bool)
    else:
        return mask * X * scale



import numpy as np


class Concat:
    def __init__(self, opset_version, **kwargs):
        self.version = opset_version
        self.axis = kwargs["axis"]

    def run(self, *inputs):
        return [np.concatenate(inputs, axis=self.axis)]



class Identity:
    def __init__(self, opset_version, **kwargs):
        self.version = opset_version

    def run(self, x):
        if type(x) == list:
            return [[v.copy() for v in x]]
        else:
            return [x.copy()]



import numpy as np


class Loop:
    def __init__(self, opset_version, **kwargs):
        self.version = opset_version
        self.body = kwargs["body"]

    def run(self, M, cond, *v_initial):
        if M == np.array(""):
            max_trip_count = np.array(np.inf, dtype=np.int64)
        else:
            max_trip_count = M.copy()

        if cond == np.array(""):
            loop_cond = np.array(True, dtype=np.bool)
        else:
            loop_cond = cond.copy()

        iter_count = np.array(0, dtype=np.int64)
        step_inputs = [v for v in v_initial]
        n = len(v_initial)

        while iter_count < max_trip_count and loop_cond:
            loop_cond, *step_outputs = self.body.run(iter_count, loop_cond, *step_inputs)
            step_inputs = [v for v in step_outputs[:n]]
            if iter_count == 0:
                scan_out = [(v,) for v in step_outputs[n:]]
            else:
                scan_out = [acc + (new,) for (acc, new) in zip(scan_out, step_outputs[n:])]
            iter_count += 1

        return step_outputs[:n] + [np.stack(vs) for vs in scan_out]



import numpy as np


class NonZero:
    def __init__(self, opset_version, **kwargs):
        self.version = opset_version

    def run(self, x):
        y = np.array(np.nonzero(x), dtype=np.int64)
        return [y]



import numpy as np


class Celu:
    def __init__(self, opset_version, **kwargs):
        self.version = opset_version
        self.alpha = kwargs.get("alpha", 1.0)

    def run(self, x):
        y = np.maximum(0, x) + np.minimum(0, self.alpha * (np.exp(x / self.alpha) - 1))
        return [y]



import numpy as np

from .utils import tensor_type_to_dtype


class EyeLike:
    def __init__(self, opset_version, **kwargs):
        self.version = opset_version
        self.dtype = kwargs.get("dtype")
        self.k = kwargs.get("k", 0)

    def run(self, x):
        sh = x.shape
        assert len(sh) == 2

        if self.dtype is None:
            dtype = x.dtype
        else:
            dtype = tensor_type_to_dtype(self.dtype)

        return [np.eye(sh[0], sh[1], k=self.k, dtype=dtype)]



from .error import RunError
from .utils import tensor_type_to_dtype


class Cast:
    def __init__(self, opset_version, **kwargs):
        self.version = opset_version
        self.to = kwargs["to"]

    def run(self, x):
        if self.version < 6:
            raise RunError("Cast", self.version)
        else:
            t = tensor_type_to_dtype(self.to)
            return [x.astype(t)]



import numpy as np


class Exp:
    def __init__(self, opset_version, **kwargs):
        self.version = opset_version
        self.consumed_inputs = kwargs.get("consumed_inputs")

    def run(self, x):
        return [np.exp(x)]



import numpy as np

from .error import RunError


class Equal:
    def __init__(self, opset_version, **kwargs):
        self.version = opset_version
        self.broadcast = kwargs.get("broadcast", 0)

    def run(self, x, y):
        if self.version > 6:
            return [np.equal(x, y)]
        else:
            raise RunError("Equal", self.version)



import numpy as np


# Ref: https://github.com/onnx/onnx/blob/bf426626378cbb715d3d0000a35f065708156cbc/onnx/onnx.proto#L266-L287
def tensor_type_to_dtype(t: int) -> np.dtype:
    mapping = [
        None,  # UNDEFINED
        np.float32,
        np.uint8,
        np.int8,
        np.uint16,
        np.int16,
        np.int32,
        np.int64,
        np.object,
        np.bool,
        np.float16,
        np.float64,
        np.uint32,
        np.uint64,
        np.complex64,
        np.complex128,
    ]

    return mapping[t]



import numpy as np


class ConstantOfShape:
    def __init__(self, opset_version, **kwargs):
        self.version = opset_version
        self.value = kwargs.get("value")

    def run(self, shape):
        n = np.prod(shape)
        if self.value is None:
            return [np.zeros(n).astype(np.float32).reshape(shape)]
        else:
            return [np.repeat(self.value, n).astype(self.value.dtype).reshape(shape)]



import numpy as np


class Expand:
    def __init__(self, opset_version, **kwargs):
        self.version = opset_version

    def run(self, x, shape):
        return [x * np.ones(shape).astype(x.dtype)]



import numpy as np


class Squeeze:
    def __init__(self, opset_version, **kwargs):
        self.version = opset_version
        self.axes = kwargs.get("axes")

    def run(self, data, axes=None):
        if self.version >= 13:
            self.axes = axes

        if self.axes is None:
            return [np.squeeze(data, axis=self.axes)]
        else:
            return [np.squeeze(data, axis=tuple(self.axes))]



from .abs import Abs  # noqa: F401
from .acos import Acos  # noqa: F401
from .acosh import Acosh  # noqa: F401
from .add import Add  # noqa: F401
from .and_ import And  # noqa: F401
from .argmax import ArgMax  # noqa: F401
from .argmin import ArgMin  # noqa: F401
from .asin import Asin  # noqa: F401
from .asinh import Asinh  # noqa: F401
from .atan import Atan  # noqa: F401
from .atanh import Atanh  # noqa: F401
from .bitshift import BitShift  # noqa: F401
from .cast import Cast  # noqa: F401
from .ceil import Ceil  # noqa: F401
from .celu import Celu  # noqa: F401
from .clip import Clip  # noqa: F401
from .compress import Compress  # noqa: F401
from .concat import Concat  # noqa: F401
from .concatfromsequence import ConcatFromSequence  # noqa: F401
from .constant import Constant  # noqa: F401
from .constantofshape import ConstantOfShape  # noqa: F401
from .cos import Cos  # noqa: F401
from .cosh import Cosh  # noqa: F401
from .depthtospace import DepthToSpace  # noqa: F401
from .dequantizelinear import DequantizeLinear  # noqa: F401
from .det import Det  # noqa: F401
from .div import Div  # noqa: F401
from .dropout import Dropout  # noqa: F401
from .dynamicquantizelinear import DynamicQuantizeLinear  # noqa: F401
from .einsum import Einsum  # noqa: F401
from .elu import Elu  # noqa: F401
from .equal import Equal  # noqa: F401
from .erf import Erf  # noqa: F401
from .error import RunError  # noqa: F401
from .exp import Exp  # noqa: F401
from .expand import Expand  # noqa: F401
from .eyelike import EyeLike  # noqa: F401
from .flatten import Flatten  # noqa: F401
from .floor import Floor  # noqa: F401
from .gather import Gather  # noqa: F401
from .gatherelements import GatherElements  # noqa: F401
from .gathernd import GatherND  # noqa: F401
from .greater import Greater  # noqa: F401
from .identity import Identity  # noqa: F401
from .leakyrelu import LeakyRelu  # noqa: F401
from .loop import Loop  # noqa: F401
from .mul import Mul  # noqa: F401
from .nonmaxsuppression import NonMaxSuppression  # noqa: F401
from .nonzero import NonZero  # noqa: F401
from .pad import Pad  # noqa: F401
from .range import Range  # noqa: F401
from .reducemax import ReduceMax  # noqa: F401
from .reducemin import ReduceMin  # noqa: F401
from .reshape import Reshape  # noqa: F401
from .round import Round  # noqa: F401
from .scatternd import ScatterND  # noqa: F401
from .shape import Shape  # noqa: F401
from .sigmoid import Sigmoid  # noqa: F401
from .slice import Slice  # noqa: F401
from .squeeze import Squeeze  # noqa: F401
from .sub import Sub  # noqa: F401
from .tile import Tile  # noqa: F401
from .topk import TopK  # noqa: F401
from .transpose import Transpose  # noqa: F401
from .unsqueeze import Unsqueeze  # noqa: F401
from .where import Where  # noqa: F401



import numpy as np


class NonMaxSuppression:
    def __init__(self, opset_version, **kwargs):
        self.version = opset_version
        self.center_point_box = kwargs.get("center_point_box", 0)

    # Ref: https://github.com/microsoft/onnxruntime/blob/55b26b69513d3972e7c0a70f0d54f3fbb8adc054/onnxruntime/core/providers/cpu/object_detection/non_max_suppression.cc#L117  # noqa: B950
    def run(self, boxes, scores, max_output_boxes_per_class=None, iou_threshold=None, score_threshold=None):

        if max_output_boxes_per_class is None:
            max_output_boxes_per_class = np.array([0], dtype=np.int64)
        if iou_threshold is None:
            iou_threshold = np.array([0], dtype=np.float32)
        if score_threshold is None:
            score_threshold = np.array([0], dtype=np.float32)

        num_batches, num_classes, _ = scores.shape
        selected_indices = list()

        for batch in range(num_batches):
            boxes_per_batch = boxes[batch, :, :]
            for cls in range(num_classes):
                scores_per_batch_per_class = scores[batch, cls, :]
                indices = np.argsort(-scores_per_batch_per_class, kind="stable")  # to pass test_nonmaxsuppression_02

                selected_indices_per_class = list()
                while len(indices) > 0 and len(selected_indices_per_class) < max_output_boxes_per_class:
                    idx0 = indices[0]
                    indices = indices[1:]

                    selected = scores_per_batch_per_class[idx0] > score_threshold
                    for idx in selected_indices_per_class:
                        if suppress_by_iou(boxes_per_batch, idx0, idx, self.center_point_box, iou_threshold):
                            selected = False
                            break

                    if selected:
                        selected_indices_per_class.append(idx0)
                        selected_indices.append([batch, cls, idx0])

        return [np.array(selected_indices).astype(np.int64)]


# Ref: https://github.com/microsoft/onnxruntime/blob/55b26b69513d3972e7c0a70f0d54f3fbb8adc054/onnxruntime/core/providers/cpu/object_detection/non_max_suppression_helper.h#L63 # noqa: B950
def suppress_by_iou(boxes, idx0, idx1, center_point_box, iou_threshold):
    box0 = boxes[idx0, :]
    box1 = boxes[idx1, :]

    if center_point_box == 0:
        x0_min = min(box0[1], box0[3])
        x0_max = max(box0[1], box0[3])
        x1_min = min(box1[1], box1[3])
        x1_max = max(box1[1], box1[3])

        intersection_x_min = max(x0_min, x1_min)
        intersection_x_max = min(x0_max, x1_max)
        if intersection_x_max <= intersection_x_min:
            return False

        y0_min = min(box0[0], box0[2])
        y0_max = max(box0[0], box0[2])
        y1_min = min(box1[0], box1[2])
        y1_max = max(box1[0], box1[2])

        intersection_y_min = max(y0_min, y1_min)
        intersection_y_max = min(y0_max, y1_max)
        if intersection_y_max <= intersection_y_min:
            return False

    else:
        w0 = box0[2]
        h0 = box0[3]
        w1 = box1[2]
        h1 = box1[3]

        x0_min = box0[0] - w0 / 2.0
        x0_max = box0[0] + w0 / 2.0
        x1_min = box1[0] - w1 / 2.0
        x1_max = box1[0] + w1 / 2.0

        intersection_x_min = max(x0_min, x1_min)
        intersection_x_max = min(x0_max, x1_max)
        if intersection_x_max <= intersection_x_min:
            return False

        y0_min = box0[1] - h0 / 2.0
        y0_max = box0[1] + h0 / 2.0
        y1_min = box1[1] - h1 / 2.0
        y1_max = box1[1] + h1 / 2.0

        intersection_y_min = max(y0_min, y1_min)
        intersection_y_max = min(y0_max, y1_max)
        if intersection_y_max <= intersection_y_min:
            return False

    intersection_area = (intersection_x_max - intersection_x_min) * (intersection_y_max - intersection_y_min)
    if intersection_area <= 0.0:
        return False

    area0 = (x0_max - x0_min) * (y0_max - y0_min)
    area1 = (x1_max - x1_min) * (y1_max - y1_min)
    union_area = area0 + area1 - intersection_area
    if area0 <= 0.0 or area1 <= 0.0 or union_area <= 0.0:
        return False

    iou = intersection_area / union_area

    return iou > iou_threshold



import numpy as np


class Floor:
    def __init__(self, opset_version, **kwargs):
        self.version = opset_version
        self.consumed_inputs = kwargs.get("consumed_inputs")

    def run(self, x):
        return [np.floor(x)]



import numpy as np


class Clip:
    def __init__(self, opset_version, **kwargs):
        self.version = opset_version
        self.consumed_inputs = kwargs.get("consumed_inputs")
        self.max = kwargs.get("max", np.inf)
        self.min = kwargs.get("min", -np.inf)

    def run(self, x, min_val=None, max_val=None):
        if self.version >= 11:
            if min_val is not None:
                self.min = min_val
            if max_val is not None:
                self.max = max_val

        return [np.clip(x, self.min, self.max)]



import numpy as np


class Compress:
    def __init__(self, opset_version, **kwargs):
        self.version = opset_version
        self.axis = kwargs.get("axis")

    def run(self, x, cond):
        return [np.compress(cond, x, axis=self.axis)]



import numpy as np


class Einsum:
    def __init__(self, opset_version, **kwargs):
        self.version = opset_version
        self.equation = kwargs["equation"]

    def run(self, *inputs):
        return [np.einsum(self.equation, *inputs)]



from .error import RunError


class Slice:
    def __init__(self, opset_version, **kwargs):
        self.version = opset_version
        self.starts = kwargs.get("starts")
        self.ends = kwargs.get("ends")
        self.axes = kwargs.get("axes")

    def run(self, x, starts=None, ends=None, axes=None, steps=None):
        if self.version >= 10:
            self.starts = starts
            self.ends = ends
            self.axes = axes

        if self.starts is None or self.ends is None:
            raise RunError("Slice", self.version)
        else:
            return [slice_ndarray(x, self.starts, self.ends, self.axes, steps)]


def slice_ndarray(x, starts, ends, axes, steps):
    if axes is None:
        ndim = len(starts)
        axes = list(range(ndim))

    idxs = [slice(0, size) for size in x.shape]

    if steps is None:
        idxs1 = [slice(s, e) for s, e in zip(starts, ends)]
    else:
        idxs1 = [slice(s, e, d) for s, e, d in zip(starts, ends, steps)]

    for i, a in enumerate(axes):
        idxs[a] = idxs1[i]

    return x[tuple(idxs)]



import math

import numpy as np


class Erf:
    def __init__(self, opset_version, **kwargs):
        self.version = opset_version

    def run(self, x):
        y = np.vectorize(math.erf)(x).astype(x.dtype)
        return [y]



import numpy as np


class Elu:
    def __init__(self, opset_version, **kwargs):
        self.version = opset_version
        self.alpha = kwargs.get("alpha", 1.0)
        self.consumed_inputs = kwargs.get("consumed_inputs")

    def run(self, x):
        y = np.clip(x, 0, np.inf) + self.alpha * (np.exp(np.clip(x, -np.inf, 0)) - 1)
        return [y]



import numpy as np


class LeakyRelu:
    def __init__(self, opset_version, **kwargs):
        self.version = opset_version
        self.alpha = kwargs.get("alpha", 0.01)
        self.consumed_inputs = kwargs.get("consumed_inputs")

    def run(self, x):
        y = np.clip(x, 0, np.inf) + np.clip(x, -np.inf, 0) * self.alpha
        return [y]



import numpy as np


class GatherND:
    def __init__(self, opset_version, **kwargs):
        self.version = opset_version
        self.batch_dims = kwargs.get("batch_dims", 0)

    def run(self, data, indices):
        return [gather_nd_impl(data, indices, self.batch_dims)]


# The following code has been copied from
# https://github.com/onnx/onnx/blob/db1a9f2388bd48e0bdde095f231a4dcc1473430a/onnx/backend/test/case/node/gathernd.py#L15-L54
# Copyrights (c) ONNX Project Contributers
# License: Apache-2.0
def gather_nd_impl(data, indices, batch_dims):
    # type: (np.ndarray, np.ndarray, int) -> np.ndarray
    # Note the data rank - will be reused multiple times later
    data_rank = len(data.shape)

    # Check input tensors' shape/rank condition
    assert indices.shape[-1] <= data_rank

    # The list of data/indice shape of batch_dims
    batch_dims_shape = []

    # The number of elements in the batch_dims for data/indice array
    batch_dims_size = 1

    # Check the shape of indice and data are identicial for batch dims.
    for i in range(batch_dims):
        batch_dims_shape.append(indices.shape[i])
        batch_dims_size *= indices.shape[i]

    # Compute output of the op as below

    # Compute shape of output array
    output_shape = (
        batch_dims_shape + list(indices.shape)[batch_dims:-1]
        if (indices.shape[-1] == data_rank - batch_dims)
        else batch_dims_shape + list(indices.shape)[batch_dims:-1] + list(data.shape)[batch_dims + indices.shape[-1] :]
    )

    # Placeholder for output data
    output_data_buffer = []

    # Flatten 'indices' to 2D array
    reshaped_indices = indices.reshape(batch_dims_size, -1, indices.shape[-1])

    # Flatten 'data' to array of shape (batch_dim_size, data.shape[batch_dimes:])
    reshaped_data = data.reshape((batch_dims_size,) + data.shape[batch_dims:])

    # gather each scalar value from 'data'
    for batch_dim in range(reshaped_indices.shape[0]):
        for outer_dim in range(reshaped_indices.shape[1]):
            gather_index = tuple(reshaped_indices[batch_dim][outer_dim])
            output_data_buffer.append(reshaped_data[(batch_dim,) + gather_index])
    return np.asarray(output_data_buffer, dtype=data.dtype).reshape(output_shape)



import numpy as np


class DynamicQuantizeLinear:
    def __init__(self, opset_version, **kwargs):
        self.version = opset_version

    def run(self, x):
        x_min = np.minimum(0, np.min(x))
        x_max = np.maximum(0, np.max(x))
        y_scale = np.array((x_max - x_min) / (255 - 0)).astype(np.float32)
        y_zero = np.clip(np.round((0 - x_min) / y_scale), 0, 255).astype(np.uint8)
        y = np.clip(np.round(x / y_scale) + y_zero, 0, 255).astype(np.uint8)

        return [y, y_scale, y_zero]



import numpy as np


class Range:
    def __init__(self, opset_version, **kwargs):
        self.version = opset_version

    def run(self, start, limit, delta):
        y = np.arange(start, limit, delta).astype(start.dtype)
        return [y]



import numpy as np


class GatherElements:
    def __init__(self, opset_version, **kwargs):
        self.version = opset_version
        self.axis = kwargs.get("axis", 0)

    def run(self, data, indices):
        return [gather_elements(data, indices, self.axis)]


# The following code has been copied from
# https://github.com/onnx/onnx/blob/db1a9f2388bd48e0bdde095f231a4dcc1473430a/onnx/backend/test/case/node/gatherelements.py#L16-L21
# Copyrights (c) ONNX Project Contributers
# License: Apache-2.0
def gather_elements(data, indices, axis=0):  # type: ignore
    data_swaped = np.swapaxes(data, 0, axis)
    index_swaped = np.swapaxes(indices, 0, axis)
    gathered = np.choose(index_swaped, data_swaped, mode="wrap")
    y = np.swapaxes(gathered, 0, axis)
    return y



import numpy as np


class ArgMin:
    def __init__(self, opset_version, **kwargs):
        self.version = opset_version
        self.axis = kwargs.get("axis", 0)
        self.keepdims = kwargs.get("keepdims", 1)
        self.select_last_index = kwargs.get("select_last_index", 0)

    def run(self, data):
        if self.version >= 12:
            if self.select_last_index:
                return [argmin_use_numpy_select_last_index(data, axis=self.axis, keepdims=self.keepdims)]
            else:
                return [argmin_use_numpy(data, axis=self.axis, keepdims=self.keepdims)]
        else:
            return [argmin_use_numpy(data, axis=self.axis, keepdims=self.keepdims)]


# The following code has been copied from
# https://github.com/onnx/onnx/blob/c4de8b7487603c4496cf6d18b3836be111d0eb90/onnx/backend/test/case/node/argmin.py#L15
# Copyrights (c) ONNX Project Contributers
# License: Apache-2.0
def argmin_use_numpy(data, axis=0, keepdims=1):  # type: (np.ndarray, int, int) -> (np.ndarray)
    result = np.argmin(data, axis=axis)
    if keepdims == 1:
        result = np.expand_dims(result, axis)
    return result.astype(np.int64)


# The following code has been copied from
# https://github.com/onnx/onnx/blob/c4de8b7487603c4496cf6d18b3836be111d0eb90/onnx/backend/test/case/node/argmin.py#L22
# Copyrights (c) ONNX Project Contributers
# License: Apache-2.0
def argmin_use_numpy_select_last_index(data, axis=0, keepdims=True):  # type: (np.ndarray, int, int) -> (np.ndarray)
    data = np.flip(data, axis)
    result = np.argmin(data, axis=axis)
    result = data.shape[axis] - result - 1
    if keepdims:
        result = np.expand_dims(result, axis)
    return result.astype(np.int64)



from .error import RunError


class BitShift:
    def __init__(self, opset_version, **kwargs):
        self.version = opset_version
        self.direction = kwargs["direction"]

    def run(self, x, y):
        if self.direction == "RIGHT":
            return [x >> y]
        elif self.direction == "LEFT":
            return [x << y]
        else:
            raise RunError("BitShift", self.version)



from .error import RunError


class Mul:
    def __init__(self, opset_version, **kwargs):
        self.version = opset_version
        self.axis = kwargs.get("axis")
        self.broadcast = kwargs.get("broadcast")
        self.consumed_inputs = kwargs.get("consumed_inputs")

    def run(self, x, y):
        if self.version > 6:
            return [x * y]
        else:
            raise RunError("Mul", self.version)



import numpy as np

from .error import RunError


class Pad:
    def __init__(self, opset_version, **kwargs):
        self.version = opset_version
        self.mode = kwargs.get("mode", "constant")
        self.pads = kwargs.get("pads")
        self.paddings = kwargs.get("paddings")
        self.value = kwargs.get("value", 0.0)

    def run(self, data, pads=None, constant_value=None):
        if self.version == 1:
            if self.paddings is None:
                raise RunError("Pad", self.version)
            pads = self.paddings
            constant_value = self.value
        elif self.version < 11:
            if self.pads is None:
                raise RunError("Pad", self.version)
            pads = self.pads
            constant_value = self.value
        else:
            if pads is None:
                raise RunError("Pad", self.version)
            if constant_value is None:
                constant_value = 0.0

        return [pad_impl(data, pads, self.mode, constant_value)]


# The following code has been copied from
# https://github.com/onnx/onnx/blob/07c494bf077e9e4a7898119f28a50585469ad4cd/onnx/backend/test/case/node/pad.py#L15-L41
# Copyrights (c) ONNX Project Contributers
# License: Apache-2.0
def pad_impl(data, raw_pads, mode, constant_values=0.0):  # type: ignore

    input_rank = data.ndim
    if input_rank * 2 != raw_pads.size:
        raise Exception("The number of elements in raw_pads should be 2 * data_rank")

    # re-order to np.pad accepted order ((x1_begin, x1_end), (x2_begin, x2_end), ...)
    pad_width = ()
    for i in range(int(raw_pads.size / 2)):
        pad_width += (((raw_pads[i], raw_pads[i + input_rank])),)  # type: ignore

    if mode == "constant":
        y = np.pad(
            data,
            pad_width=pad_width,
            mode=mode,
            constant_values=constant_values,
        )
        return y

    y = np.pad(
        data,
        pad_width=pad_width,
        mode=mode,
    )

    return y



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['onnion_runtime']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.19.0,<2.0.0']

setup_kwargs = {
    'name': 'onnion-rt',
    'version': '0.3.0',
    'description': 'run onnx with only numpy',
    'long_description': '# onnion-rt\n\nNote: This software includes [the work](https://github.com/onnx/onnx) that is distributed in the [Apache License 2.0](https://www.apache.org/licenses/LICENSE-2.0.html).\n\n## Installation\n\n```\n$ pip3 install onnion-rt\n```\n\n## Usage\nSee [tutorial](https://github.com/Idein/onnion/tree/master#tutorial).\n\n## Development Guide\n\n```\n$ poetry install\n```\n\n### How to support new operators\n\n1. Add `onnion_runtime/xxx.py`\n  - An onnx operator `Xxx` must correspond to a class `Xxx`.\n  - A class `Xxx` must implement `__init__` and `run` methods.\n  - The parameters of the `__init__` methods must be `self`, `opset_version`, and `kwargs`.\n  - The attributes of the operator must be passed through the `kwargs` of the `__init__` method.\n    - Get the required attributes by `kwargs[\'attr_name\']`.\n    - Get the optional attributes by `kwargs.get(\'attr_name\', default_value)`.\n  - The inputs of the operator must be passed through the arguments of the `run` method.\n  - The `run` method must return the list of `np.array`.\n2. Add `from .xxx import Xxx # noqa: F401` to `onnion_runtime/__init__.py`\n3. Update "Supported Operators" in `README.md`\n4. Add `tests/test_xxx.py`\n5. Run tests `poetry run pytest -v`\n6. Format and lint `poetry run pysen run format && poetry run pysen run lint`\n\n## Supported Operators\nThis runtime supports only below operators.\n\n- Abs\n- Acos\n- Acosh\n- Add\n  - must be from opsetversion >= 7\n- And\n  - must be from opsetversion >= 7\n- ArgMax\n- ArgMin\n- Asin\n- Asinh\n- Atan\n- Atanh\n- BitShift\n- Cast\n  - must be from opsetversion >= 6\n- Ceil\n- Celu\n- Clip\n- Compress\n- Concat\n- ConcatFromSequence\n- Constant\n- ConstantOfShape\n- Cos\n- Cosh\n- DepthToSpace\n- DequantizeLinear\n- Det\n- Div\n  - must be from opsetversion >= 7\n- Dropout\n- DynamicQuantizeLinear\n- Einsum\n- Elu\n- Equal\n  - must be from opsetversion >= 7\n- Erf\n- Exp\n- Expand\n- EyeLike\n- Flatten\n- Floor\n- Gather\n- GatherElements\n- GahterND\n- Greater\n  - must be from opsetversion >= 7\n- Identity\n- LeakyRelu\n- Loop\n- Mul\n  - must be from opsetversion >= 7\n- NonMaxSuppression\n- NonZero\n- Pad\n- Range\n- ReduceMax\n- ReduceMin\n- Reshape\n- Round\n- ScatterND\n- Shape\n- Sigmoid\n- Slice\n- Squeeze\n- Sub\n  - must be from opsetversion >= 7\n- Tile\n  - must be from opsetversion >= 6\n- TopK\n- Transpose\n- Unsqueeze\n- Where\n',
    'author': 'Idein Inc.',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Idein/onnion/tree/master/runtime',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



import numpy as np


class Sigmoid:
    def __init__(self, opset_version, **kwargs):
        self.version = opset_version
        self.consumed_inputs = kwargs.get("consumed_inputs")

    def run(self, x):
        return [1.0 / (1.0 + np.exp(np.negative(x)))]



import numpy as np

from .error import RunError


class Constant:
    def __init__(self, opset_version, **kwargs):
        self.version = opset_version
        self.sparse_value = kwargs.get("sparse_value")
        self.value = kwargs.get("value")
        self.value_float = kwargs.get("value_float")
        self.value_floats = kwargs.get("value_floats")
        self.value_int = kwargs.get("value_int")
        self.value_ints = kwargs.get("value_ints")
        self.value_string = kwargs.get("value_string")
        self.value_strings = kwargs.get("value_strings")

    def run(self):
        if self.version < 11:
            if self.value is not None:
                return [self.value]
            else:
                raise RunError("Constant", self.version)
        elif self.version == 11:
            if self.value is not None:
                return [self.value]
            elif self.sparse_value is not None:
                return [self.sparse_value]
            else:
                raise RunError("Constant", self.version)
        else:
            if self.value is not None:
                return [self.value]
            elif self.sparse_value is not None:
                return [self.sparse_value]
            elif self.value_float is not None:
                return [np.array(self.value_float).astype(np.float32)]
            elif self.value_floats is not None:
                return [np.array(self.value_floats).astype(np.float32)]
            elif self.value_int is not None:
                return [np.array(self.value_int).astype(np.int64)]
            elif self.value_ints is not None:
                return [np.array(self.value_ints).astype(np.int64)]
            elif self.value_string is not None:
                return [np.array(self.value_string)]
            elif self.value_strings is not None:
                return [np.array(self.value_strings)]
            else:
                raise RunError("Constant", self.version)



import numpy as np

from .error import RunError


class Reshape:
    def __init__(self, opset_version, **kwargs):
        self.version = opset_version
        self.allowzero = kwargs.get("allowzero", 0)
        self.consumed_inputs = kwargs.get("consumed_inputs")
        self.shape = kwargs.get("shape")

    def run(self, data, shape=None):
        if self.version >= 5:
            self.shape = shape

        if self.shape is None:
            raise RunError("Reshape", self.version)

        return [reshape_reference_implementation(data, self.shape, allowzero=self.allowzero)]


# The following code has been copied from
# https://github.com/onnx/onnx/blob/a71135c16c6c8f95fe3b0df2144955f872cb7cda/onnx/backend/test/case/node/reshape.py#L15.
# Copyrights (c) ONNX Project Contributers
# License: Apache-2.0
def reshape_reference_implementation(data, shape, allowzero=0):  # type: (np.ndarray, np.ndarray, int) -> np.ndarray
    # replace zeros with corresponding dim size
    # we need to do this because np.reshape doesn't support 0 by default unless 'allowzero' is set
    new_shape = np.copy(shape)
    if allowzero == 0:
        zeros_index = np.where(shape == 0)
        new_shape[zeros_index] = np.array(data.shape)[zeros_index]
    reshaped = np.reshape(data, new_shape)
    return reshaped



import numpy as np

from .error import RunError


class DepthToSpace:
    def __init__(self, opset_version, **kwargs):
        self.version = opset_version
        self.blocksize = kwargs["blocksize"]
        self.mode = kwargs.get("mode", "DCR")

    def run(self, x):
        assert len(x.shape) == 4
        n, c, h, w = x.shape

        if self.mode == "DCR":
            tmp = np.reshape(x, [n, self.blocksize, self.blocksize, c // (self.blocksize ** 2), h, w])
            tmp = np.transpose(tmp, [0, 3, 4, 1, 5, 2])
            y = np.reshape(tmp, [n, c // (self.blocksize ** 2), h * self.blocksize, w * self.blocksize])
            return [y]
        elif self.mode == "CRD":
            tmp = np.reshape(x, [n, c // (self.blocksize ** 2), self.blocksize, self.blocksize, h, w])
            tmp = np.transpose(tmp, [0, 1, 4, 2, 5, 3])
            y = np.reshape(tmp, [n, c // (self.blocksize ** 2), h * self.blocksize, w * self.blocksize])
            return [y]
        else:
            raise RunError("DepthToSpace", self.version)



import numpy as np

from .error import RunError


class Unsqueeze:
    def __init__(self, opset_version, **kwargs):
        self.version = opset_version
        self.axes = kwargs.get("axes")

    def run(self, data, axes=None):
        if self.version >= 13:
            self.axes = axes

        if self.axes is None:
            raise RunError("Unsqueeze", self.version)

        result = data.copy()
        for a in sorted(self.axes):
            result = np.expand_dims(result, axis=a)

        return [result]



import numpy as np


class ConcatFromSequence:
    def __init__(self, opset_version, **kwargs):
        self.version = opset_version
        self.axis = kwargs["axis"]
        self.new_axis = kwargs.get("new_axis", 0)

    def run(self, inputs):
        if self.new_axis == 0:
            return [np.concatenate(inputs, axis=self.axis)]
        else:
            return [np.stack(inputs, axis=self.axis)]
